import functools
import time

from requests.exceptions import ConnectionError

from robot.api import logger

def retry_on_exception(func):
    
    @functools.wraps(func)
    def retry_helper(self, *args, **kwargs):
        current_retry = 0

        local_kwargs = kwargs.copy()

        current_delay = local_kwargs.get('delay', self._delay)
        backoff_factor = local_kwargs.get('backoff_factor', self._backoff_factor)
        max_retries = local_kwargs.get('max_retries', self._max_retries)

        args_repr = [repr(a) for a in args]
        kwargs_repr = [f"{k}={v!r}" for k, v in local_kwargs.items()]
    
        signature = ", ".join(args_repr + kwargs_repr)
    
        logger.debug(f"Calling retry for {func.__name__}({signature})")

        logger.debug(f"_max_retries: {max_retries}")
        logger.debug(f"_backoff_factor: {backoff_factor}")
        logger.debug(f"_delay: {current_delay}")
        

        # delete the max_retries, delay and backoff_factor from the kwargs
        if 'max_retries' in local_kwargs:
            #del local_kwargs['max_retries']
            pass

        if 'delay' in local_kwargs:
            #del local_kwargs['delay']
            pass

        if 'backoff_factor' in local_kwargs:
            #del local_kwargs['backoff_factor']
            pass

        while True:
            try:
                value = func(self, *args, **local_kwargs)
    
                logger.debug(f"{func.__name__!r} returned {value!r}")
            
                return value
                
            except ConnectionError as e:
                time.sleep(current_delay)
                current_retry = current_retry + 1
                current_delay = current_delay * backoff_factor
                
                if current_retry > max_retries:
                    raise Exception(f"Calling {func.__name__} with max_retries {max_retries} without success.") from e

                logger.info(f"[{func.__name__}] Retry count: {current_retry} of {max_retries}; delay: {current_delay} and backoff_factor: {backoff_factor}")

    return retry_helper



import docker
import time

from robot.api import logger


def str2bool(v):
    return v.lower() in ("yes", "true", "t", "1")


class DockerHandler:

    IMAGE_NAME = '5minds/engine'

    def __init__(self, **kwargs):

        params = {
            'timeout': 120
        }

        self._client = docker.from_env(**params)
        self._name = kwargs.get(
            'container_name', 'robotframework_5minds-engine')
        self._auto_remove = str2bool(kwargs.get('auto_remove', True))
        self._image_name = kwargs.get('image_name', DockerHandler.IMAGE_NAME)

        self._delay = kwargs.get('delay', 5.0)

    def start(self):
        logger.console(f"Starting engine within a docker container '{self._image_name}' ...")

        args = {
            'detach': True,
            'publish_all_ports': True,
            'name': self._name,
            'auto_remove': self._auto_remove,
        }

        if self.find_container():
            if self._container.status == 'exited':
                self._container.restart()
        else:
            self._container = self._client.containers.run(
                self._image_name, **args)

        time.sleep(float(self._delay))

        container = self._client.containers.get(self._container.id)

        key = list(container.attrs['NetworkSettings']['Ports'].keys())[0]

        self._port = container.attrs['NetworkSettings']['Ports'][key][0]['HostPort']

        logger.console(f"Starting engine within a docker container '{self._image_name}' ... done")


    def get_engine_url(self):
        engine_url = f"http://localhost:{self._port}"

        return engine_url

    def find_container(self):
        containers = self._client.containers.list(
            all=True, filters={'name': self._name})

        if len(containers) > 0:
            self._container = containers[0]
            found = True
        else:
            found = False

        return found

    def shutown(self, prune=False):
        self._container.stop()

        if prune:
            self._client.containers.prune()

    def report(self):
        print(f"container id: {self._container.id}")
        print(f"container attr: {self._port}")


def main():
    handler = DockerHandler(auto_remove=False)

    handler.start()
    handler.report()
    handler.shutown()


if __name__ == '__main__':
    main()




from atlas_engine_client.core.api.client import DeployResults

from robot.api import logger

from ._retry_helper import retry_on_exception


class DeployKeyword:

    def __init__(self, client, **kwargs):
        self._client = client

    @retry_on_exception
    def deploy_processmodel(self, pathname, exit_on_fail: bool = True, overwrite_existing: bool = True, **kwargs) -> DeployResults:
        deploy_results = self._client.process_defintion_deploy_by_pathname(
            pathname, exit_on_fail, overwrite_existing)

        for deploy_result in deploy_results.deployed_files:
            logger.info(
                f"Deploy '{deploy_result.filename}' successfully=='{deploy_result.deployed}'")

        return deploy_results




from .deploy_keyword import DeployKeyword
from .empty_task_keyword import EmptyTaskKeyword
from .engine_keyword import EngineKeyword
from .external_task_keyword import ExternalTaskKeyword
from .manual_task_keyword import ManualTaskKeyword
from .process_instance_keyword import ProcessInstanceKeyword
from .send_keyword import SendKeyword
from .start_keyword import StartKeyword
from .user_task_keyword import UserTaskKeyword


import json
import time
from typing import Dict, Any
from dataclasses import fields

from atlas_engine_client.core.api import FlowNodeInstancesQuery
from atlas_engine_client.core.api import FlowNodeInstanceResponse

from robot.api import logger

from ._fields_helper import filter_kwargs_for_dataclass
from ._retry_helper import retry_on_exception


class ProcessInstanceKeyword:

    def __init__(self, client, **kwargs):
        self._client = client

        self._max_retries = kwargs.get('max_retries', 5)
        self._backoff_factor = kwargs.get('backoff_factor', 2)
        self._delay = kwargs.get('delay', 0.1)

    @retry_on_exception
    def get_processinstance(self, **kwargs) -> FlowNodeInstanceResponse:
        return self._get_processinstance(**kwargs)

    @retry_on_exception
    def get_processinstance_result(self, **kwargs) -> Dict[str, Any]:
        result = self._get_processinstance(**kwargs)

        if result and len(result.tokens) > 0:
            payload = result.tokens[0]['payload']
            if payload is not None:
                try:
                    logger.info(f"type(payload) {type(payload)}")
                    if type(payload) in [str, bytes]:
                        payload = json.loads(payload)
                    else:
                        pass
                except json.decoder.JSONDecodeError:
                    payload = {}
        else:
            payload = {}


        return payload

    def _get_processinstance(self, **kwargs) -> FlowNodeInstanceResponse:

        query_dict = {
            'state': 'finished',
            'limit': 1,
            'flow_node_type': 'bpmn:EndEvent',
        }

        current_retry = 0
        current_delay = float(kwargs.get('delay', self._delay))
        backoff_factor = float(kwargs.get('backoff_factor', self._backoff_factor))
        max_retries = int(kwargs.get('max_retries', self._max_retries))

        local_kwargs = filter_kwargs_for_dataclass(FlowNodeInstancesQuery, kwargs)

        query_dict.update(**local_kwargs)

        while True:

            query = FlowNodeInstancesQuery(**query_dict)

            flow_node_instances = self._client.flow_node_instance_get(query)

            if len(flow_node_instances) == 1:
                flow_node_instance = flow_node_instances[0]
            else:
                flow_node_instance = None

            if flow_node_instance:
                break
            else:
                time.sleep(current_delay)
                current_retry = current_retry + 1
                current_delay = current_delay * backoff_factor
                if current_retry > max_retries:
                    break
                logger.info(f" Retry count: {current_retry} of {max_retries}; delay: {current_delay} and backoff_factor: {backoff_factor}")

        return flow_node_instance
